/*
 * This is a parent process that fork/exec's a child.  It tries to
 * find possible problems with releasing names when a parent dies with
 * an SRR name attached while the child continues to run.  The name
 * would be left attached to a dead process if the SRR module cannot
 * figure out that the parent has died.
 *
 * Here's the problem:
 *
 * - Parent (P) registers a name, which sets up a FD to connect to the
 * SRR module.
 *
 * - P starts a thread.  When the first thread (T) is started, the
 * pthread library starts a helper thread (H) that starts the real
 * thread, T.  T inherits the file descriptors of the mainline,
 * including FD.
 *
 * - T fork/execs a child process C, which inherits FD from T.
 *
 * - C calls setsid to detach from P.
 *
 * - P dies.  Therefore, T dies.  FD is closed by H on behalf of T,
 * but this also appears to close FD on behalf of P.  P never closes
 * FD explicitly on exit, so the SRR module does not see a close by P,
 * and does not release FD and its associated names yet.
 *
 * - C dies later.  When C closes FD, SRR sees the final release on FD
 * and releases all associated names, even though FD was closed by the
 * wrong process.
 *
 * - If C does not die, we get a dangling FD with names associated
 * with a dead process.
 */

#include	<stdio.h>
#include	<stdlib.h>
#include	<string.h>
#include	<errno.h>
#include	<unistd.h>
#include	<pthread.h>
#include	<srr.h>

int Survive = 0;

void startchild (void* data)
{
	int		fk;
	
	if ((fk = fork()) == 0)
	{
		if (execl ("./name2", NULL) == -1)
		{
			printf ("Child could not execute: %s\n", strerror(errno));
			exit (-1);
		}
	}
	else if (fk == -1)
	{
		printf ("Fork failed: %s\n", strerror(errno));
		exit (-1);
	}
	else
	{
		printf ("Fork child: %d\n", fk);
	}
	sleep (10);
	printf ("Error: Launch thread exited normally (it should be killed).\n");
}

int main (int argc, char** argv)
{
	int			iret;
	pthread_t	tid;

	if (argc > 1)
		Survive = 1;

	if (Survive)
		SrrSurviveExec();
	
	if (SrrNameAttach (0, "name1") == -1)
	{
		printf ("Parent failed to attach name: name1: %s\n", strerror(errno));
		return (-1);
	}
	iret = pthread_create(&tid, NULL, (void*)&startchild, NULL);
	sleep (4);
	return (0);
}
