/*
 * srr.h: interface to send/receive/reply IPC library calls.
 *
 *  Linux Loadable Module Implementation of Send/Receive/Reply
 *	message passing and related functions, similar to what is used
 *  by QNX4.
 *
 *	QNX is a trademark of QNX Software Systems Ltd. <www.qnx.com>
 *
 * Copyright (C) 2000 Cogent Real-Time Systems Inc.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * The library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; see the file COPYING.LIB.  If not,
 * write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 * $Id: srr.h 3023 2006-10-13 21:50:19Z andrew $
 */

#ifndef SRR_SRR_H
#define SRR_SRR_H

#ifdef __cplusplus
extern "C" {
#endif

#ifndef __KERNEL__

#include <signal.h>
#include <sys/times.h>
#include <sys/types.h>
#include <time.h>

#ifndef getnid
#define getnid()	0
#endif
  
/* There is a real problem with the infinite supply of Linux
   kernels that might or might not define struct itimerspec.
   Define this if your kernel headers do not define it.  */
// #define NO_KERNEL_ITIMERSPEC

#endif /* not __KERNEL__ */

#include "qnxtypes.h"
#include "sys/sendmx.h"

/*
 * mx record for multi-part messages
 * If you comment this definition for _mxfer_entry, and uncomment the
 * one below, you can send messages in mxfer functions larger than
 * 64K.  This will, unfortunately, break any programs that are
 * compiled with the smaller mxfer structure, so you have to
 * recompile everything, including software you bought from Cogent.
 */

struct _mxfer_entry {
	void	*addr;
	short	unsigned	mxfer_len;
	};

// allow huge messages
//struct _mxfer_entry {
//	void	*addr;
//	int	unsigned	mxfer_len;
//	};

#define _setmx(mx,msg,len) \
	{ (mx)->addr = msg; \
	  (mx)->mxfer_len = len; }
	
/*
 * pre-2.2.x kernels don't have sigevent defined, so define it
 * ourselves, conditionally based on one of the defines used
 * with the sigevent structure.
 */
#ifndef SIGEV_SIGNAL
struct sigevent
{
	int sigev_signo;
};
#endif

/* This is defined in <linux/time.h>, but whoever wrote that
 * file had no intention that it be used outside the kernel
 */

#if (!(defined(__USE_POSIX199309) && defined(_TIME_H)) && !defined(ITIMER_REAL)) || defined(NO_KERNEL_ITIMERSPEC)
struct  itimerspec {
	struct  timespec it_interval;    /* timer period */
	struct  timespec it_value;       /* timer expiration */
};
#endif

#define CLOCK_REALTIME  0   // default POSIX clock, must be present on all systems

typedef struct receiveTimeoutStruct {
  pid_t pid;
  long time;
} receiveTimeoutStruct;

typedef struct sendTimeoutStruct {
  size_t rsize;
  long time;
} sendTimeoutStruct;

// flags for timer_settime() - along with my guess as to their providence
// Some Linux kernels, in no particular order of version number, define
// some of these flags.
#ifndef TIMER_ABSTIME
#define TIMER_ABSTIME           0x0001  // POSIX
#endif

#define TIMER_ADDREL            0x0100  // POSIX?
#define TIMER_PRESERVE_EXEC     0x0200  // QNX?
#define TIMER_AUTO_RELEASE      0x0400  // QNX?

// Non-Qnx and module-specific functions.
int SrrUsePthreads(int enable);
int SrrReg(void);
	// EBADF  - "/dev/srripc" exists, but the module is not loaded
	// ENOENT - "/dev/srripc" doesn't exist
	// Note: SrrReg() is called by all the API functions, thus these
	//    errors can be returned from any of them.
int SrrUnreg(void);
int SrrUnload(void);
int SrrFd(void);
int SrrSurviveExec (void);
int SrrDebug(int level);
int SrrKill(pid_t pid, int signo);

// SRR Send/Receive/Reply message passing.

int SrrSend(pid_t pid, void* smsg, void* rmsg, size_t ssize, size_t rsize);
	// EFAULT - process would have incurred a segv, buffers invalid or too small
	// EINTR - interrupted by a signal
	// EINVAL - message length invalid
	// ENOMEM - insufficient memory to grow buffer
	// ESRCH - process pid does not exist
sendTimeoutStruct SrrSendTimeout(pid_t pid, void* smsg, void* rmsg, size_t ssize, size_t rsize, unsigned long time);
	// EFAULT - process would have incurred a segv, buffers invalid or too small
	// EINTR - interrupted by a signal
	// EINVAL - message length invalid
	// ENOMEM - insufficient memory to grow buffer
	// ESRCH - process pid does not exist
pid_t SrrReceive(pid_t pid, void* msg, size_t *size);
	// EFAULT
	// EINTR
	// ESRCH
receiveTimeoutStruct SrrRecieveTimeout(pid_t pid, void* msg, size_t *size, unsigned long time);	// EFAULT
	// EINTER
	// ESRCH
pid_t SrrCreceive(pid_t pid, void* msg, size_t *size);
	// EFAULT
	// EINTR
	// ESRCH
	// ENOMSG
int SrrReply(pid_t pid, void* msg, size_t size);
	// EFAULT
	// EINTR
	// EINVAL
	// ENOMEM
	// ESRCH

/* 2005/09/05 - Len Meakin */
int SrrRelay(pid_t src, pid_t tgt);
	// ESRCH
	
/* 2000/02/17 - Alessandro Sala */
int SrrReadmsg(pid_t pid, size_t offset, void* msg, size_t size);
	// EFAULT
	// EINTR
	// ESRCH

/* 2000/03/29 - Alessandro Sala */
int SrrWritemsg(pid_t pid, size_t offset, void* msg, size_t size);
	// EFAULT
	// EINTR
	// ESRCH
	// ENOMEM
/*
** Begin MX
** John Green 2000/10/9
*/
int SrrSendmx(pid_t pid, size_t sparts, size_t rparts,
			  struct _mxfer_entry *smsg, struct _mxfer_entry *rmsg);
        // EFAULT - process would have incurred a segv, buffers
        // invalid or too small
        // EINTR - interrupted by a signal
        // EINVAL - message length invalid
        // ENOMEM - insufficient memory to grow buffer
        // ESRCH - process pid does not exist
sendTimeoutStruct SrrSendmxTimeout(pid_t pid, size_t sparts, size_t rparts,
			  struct _mxfer_entry *smsg, struct _mxfer_entry *rmsg, unsigned long time);
        // EFAULT - process would have incurred a segv, buffers
        // invalid or too small
        // EINTR - interrupted by a signal
        // EINVAL - message length invalid
        // ENOMEM - insufficient memory to grow buffer
        // ESRCH - process pid does not exist
pid_t SrrReceivemx(pid_t pid, size_t parts, struct _mxfer_entry *msgmx);
        // EFAULT
        // EINTR
        // ESRCH
receiveTimeoutStruct SrrReceivemxTimeout(pid_t pid, size_t parts, struct _mxfer_entry *msgmx, unsigned long time);
        // EFAULT
        // EINTR
        // ESRCH
pid_t SrrCreceivemx(pid_t pid, size_t parts, struct _mxfer_entry *msgmx);
        // EFAULT
        // EINTR
        // ESRCH
        // ENOMSG
int SrrReplymx(pid_t pid, size_t parts, struct _mxfer_entry *msgmx);
        // EFAULT
        // EINTR
        // EINVAL
        // ENOMEM
        // ESRCH
int SrrReadmsgmx(pid_t pid, size_t offset, size_t parts,
				 struct _mxfer_entry* msg);
        // EFAULT
        // EINTR
        // ESRCH
int SrrWritemsgmx(pid_t pid, size_t offset, size_t  parts,
				  struct _mxfer_entry *msgmx);
        // EFAULT
        // EINTR   
/*
** End MX
*/

// Srr name services.
int SrrNameAttach(nid_t, const char* name);
	// EBUSY - name exists
	// EAGAIN - name space used up
	// ENOMEM - no memory to fulfill request
	//	note: not a (documented) return value in QNX
int SrrNameDetach(nid_t, int name_id);
	// EINVAL - name does not exist, or not owned by you
pid_t SrrNameLocate(nid_t, const char* name, unsigned, unsigned* copies);
	// ESRCH - no process with this name

// Srr proxies.
pid_t SrrProxyAttach(pid_t pid, char* data, int nbytes, int priority);
	// EAGAIN - no free process entries
	// EINVAL - size of proxy is greater than _SRR_PROXY_SIZE_MAX
	// ENOMEM - not enough memory to buffer message
	// ESRCH - pid does not exist

int SrrProxyDetach(pid_t proxy);
	// EPERM - you are not owner of the proxy
	// ESRCH - proxy does not exist

pid_t SrrProxyTrigger(pid_t pid);
	// ESRCH - pid does not exist
	// EINVAL - pid is not a proxy
	// ? (forgot to look these up in the QNX docs)


// Srr process flags

#define _PPF_IMMORTAL       0x00000001L
#define _PPF_INFORM         0x00000002L
#define _PPF_FIXED          0x00000004L
#define _PPF_PRIORITY_REC   0x00000010L
#define _PPF_PRIORITY_FLOAT 0x00000020L
#define _PPF_NOCLDSTOP      0x00000040L
#define _PPF_SIGCATCH       0x00000080L
#define _PPF_SIGMASKALL     0x00000100L
#define _PPF_SERVER         0x00000200L
#define _PPF_RDOS           0x00000400L
#define _PPF_WAS_32BIT      0x00008000L     /*  RO */
#define _PPF_32BIT          0x00020000L     /*  RO */
#define _PPF_FLAT           0x00040000L     /*  RO */
#define _PPF_XIP            0x00080000L     /*  RO */
#define _PPF_VID            0x00100000L     /*  RO */
#define _PPF_MID            0x00200000L     /*  RO */
#define _PPF_VMID           0x00400000L     /*  RO */
#define _PPF_EXECING        0x01000000L     /*  RO */
#define _PPF_LOADING        0x02000000L     /*  RO */
#define _PPF_TERMING        0x04000000L     /*  RO */
#define _PPF_TO_BE_HELD     0x08000000L     /*  RO */
#define _PPF_NOZOMBIE       0x10000000L     /*  RO */
#define _PPF_SLEADER        0x20000000L     /*  RO */
#define _PPF_FORKED         0x40000000L     /*  RO */

// Process flags we added ourselves

#define _PPF_POLLING		0x00010000L		/*  RO */
#define _PPF_KBUF_WRITTEN 0x80000000L	/*  RO */

int SrrPflags (unsigned short flags, unsigned short mask, void* d2, void* d1);

// SRR System Messages

struct _sysmsg_hdr_reply
{
	int		status;
};

struct _sysmsg_hdr
{
	short	subtype;
	int		status;
};

#define _SYSMSG_SUBTYPE_DEATH	0

// We need a getpid for threads.
int SrrGetpid (void);

// Srr VC-related calls (trivially implemented to work on a single machine)

pid_t SrrProxyRemAttach (nid_t, pid_t pid);
int   SrrProxyRemDetach (nid_t, pid_t pid);

pid_t SrrVcAttach (nid_t, pid_t pid, int size, int flags);
int   SrrVcDetach (pid_t pid);

// POSIX.4 timers with QNX extensions (which are not implemented
// as of Linux 2.2)

/* We use srr_* versions of timer and clock functions to avoid
   problems with the sporadic and incomplete implementations
   that some kernels (or at least kernel headers) appear to
   contain.  We do not want to use even complete implementations
   of the timer_* functions because the QNX semantics extend a
   struct sigevent to include a PROXY */

srr_timer_t	SrrTimerCreate (srr_clockid_t clock, struct sigevent *event);
int		SrrTimerDelete (srr_timer_t tid);
int		SrrTimerGettime (srr_timer_t tid, struct itimerspec *value);
int		SrrTimerSettime (srr_timer_t tid, int which,
						 const struct itimerspec *val,
						 struct itimerspec *oval);

#define timer_create(c,e)	SrrTimerCreate(c,e)
#define timer_delete(t)		SrrTimerDelete(t)
#define timer_gettime(t,v)	SrrTimerGettime(t,v)
#define timer_settime(t,w,v,o)	SrrTimerSettime(t,w,v,o)

// POSIX clock API
int SrrClockSettime (srr_clockid_t clock, struct timespec *tp);
int SrrClockGettime (srr_clockid_t clock, struct timespec *tp);
int SrrClockGetres  (srr_clockid_t clock, struct timespec *tp);

#define clock_settime(c,t)	SrrClockSettime(c,t)
#define clock_gettime(c,t)	SrrClockGettime(c,t)
#define clock_getres(c,t)	SrrClockGetres(c,t)

#define timer_t				srr_timer_t
#define clockid_t			srr_clockid_t

// QNX4.2x Mappings
#define SendTimeout(pid, smsg, rmsg, ssize, rsize, time) \
  SrrSendTimeout(pid, smsg, rmsg, ssize, rsize, time)
#define Send(pid,smsg,rmsg,ssize,rsize) \
	(SrrSend(pid,smsg,rmsg,ssize,rsize) == -1 ? -1 : 0)
/* Receive is not a macro.  Indirection level on size changes */

//#define Receive(pid,msg,size)		SrrReceive(pid,msg,size)
//#define Creceive(pid,msg,size)	SrrCreceive(pid,msg,size)
extern pid_t Receive(pid_t pid, void* msg, size_t size);
extern pid_t Creceive(pid_t pid, void* msg, size_t size);
extern receiveTimeoutStruct ReceiveTimeout(pid_t pid, void* msg, size_t size, unsigned long time);
/* Sendmx, Receivemx are not macros.  SRR module changes the length
   fields of MX records. */
/* Changed these from macros to funtions to support the fact that QNX4
   does not alter the sizes in the mx records on return.  (AT)*/
/*
#define Sendmx(pid,sparts,rparts,smsg,rmsg) \
	(SrrSendmx(pid,sparts,rparts,smsg,rmsg) == -1 ? -1 : 0)
#define Receivemx(pid,parts,msgmx) \
	SrrReceivemx(pid,parts,msgmx)
#define Creceivemx(pid,parts,msgmx) \
	SrrCreceivemx(pid,parts,msgmx)
*/
pid_t Receivemx (pid_t pid, size_t parts, struct _mxfer_entry *msgmx);
receiveTimeoutStruct ReceivemxTimeout (pid_t pid, size_t parts, 
    struct _mxfer_entry *msgmx, unsigned long time);
pid_t Creceivemx (pid_t pid, size_t parts, struct _mxfer_entry *msgmx);
int Sendmx (pid_t pid, size_t sparts, size_t rparts,
			struct _mxfer_entry *smsg, struct _mxfer_entry *rmsg);
sendTimeoutStruct SendmxTimeout (pid_t pid, size_t sparts, size_t rparts,
    struct _mxfer_entry *smsg, struct _mxfer_entry *rmsg, unsigned long time);
#define Reply(pid,msg,size) \
	SrrReply(pid,msg,size)
#define Readmsg(pid,offset,msg,size) \
	SrrReadmsg(pid,offset,msg,size)
#define Writemsg(pid,offset,msg,size) \
	SrrWritemsg(pid,offset,msg,size)
#define Replymx(pid,parts,msgmx) \
	SrrReplymx(pid,parts,msgmx)
#define Readmsgmx(pid,offset,parts,msg) \
	SrrReadmsgmx(pid,offset,parts,msg)
#define Writemsgmx(pid,offset,parts,msgmx) \
	SrrWritemsgmx(pid,offset,parts,msgmx)
#define Trigger(pid) \
	SrrProxyTrigger(pid)
#define Relay(src,tgt) \
	SrrRelay(src,tgt)
	
#define qnx_name_attach(nid,name) \
	SrrNameAttach(nid,name)
#define qnx_name_detach(nid,name_id) \
	SrrNameDetach(nid,name_id)
#define qnx_name_locate(nid,name,size,copies) \
	SrrNameLocate(nid,name,size,copies)
#define qnx_proxy_attach(pid,data,nbytes,priority) \
	SrrProxyAttach(pid,data,nbytes,priority)
#define qnx_proxy_detach(proxy) \
	SrrProxyDetach(proxy)
#define qnx_pflags(flags,mask,d2,d1) \
	SrrPflags(flags,mask,d2,d1)

#define qnx_proxy_rem_attach(nid,pid) \
	SrrProxyRemAttach(nid,pid)
#define qnx_proxy_rem_detach(nid,pid) \
	SrrProxyRemDetach(nid,pid)
#define qnx_vc_attach(nid,pid,size,flags) \
	SrrVcAttach(nid,pid,size,flags)
#define qnx_vc_detach(pid) \
	SrrVcDetach(pid)

/* A whole section for IRQ handler within the module */
#define SRRIRQ_IMMEDIATE	1
#define SRRIRQ_INTERRUPT	0
#define SRRIRQ_MAXCMD		16	/* max bytes in a irq operation, incl opcode */
#define REG_A				0
#define REG_B				1
#define REG_C				2
#define REG_D				3

#if !defined(NO_IRQ) && !defined(__KERNEL__)
#include	<stdint.h>

int Srr_Mkop_mapa(char* buf, uint8_t adbuf, uint32_t offset, uint32_t size);
int Srr_Mkop_maps(char* buf, uint8_t stbuf, uint32_t size);
int Srr_Mkop_mapp(char* buf, uint16_t ioport, uint16_t size);
int Srr_Mkop_ldreg(char* buf, uint8_t reg, uint32_t value);
int Srr_Mkop_ldregas(char* buf, uint8_t opcode, uint8_t reg, uint8_t bufno,
					 uint32_t offset, uint8_t size);
int Srr_Mkop_ldrega (char* buf, uint8_t reg, uint8_t bufno,
					 uint32_t offset, uint8_t size);
int Srr_Mkop_ldregs (char* buf, uint8_t reg, uint8_t bufno,
					 uint32_t offset, uint8_t size);
int Srr_Mkop_ldregp (char* buf, uint8_t reg, uint8_t port_reg, uint8_t size);
int Srr_Mkop_stregas (char* buf, uint8_t opcode, uint8_t reg, uint8_t bufno,
					  uint32_t offset, uint8_t size);
int Srr_Mkop_strega (char* buf, uint8_t opcode, uint8_t reg, uint8_t bufno,
					 uint32_t offset, uint8_t size);
int Srr_Mkop_stregs (char* buf, uint8_t opcode, uint8_t reg, uint8_t bufno,
					 uint32_t offset, uint8_t size);
int Srr_Mkop_stregp (char* buf, uint8_t reg, uint8_t port_reg, uint8_t size);
int Srr_Mkop_regbin (char* buf, uint8_t opcode, uint8_t reg1, uint8_t reg2);
int Srr_Mkop_add (char* buf, uint8_t reg1, uint8_t reg2);
int Srr_Mkop_sub (char* buf, uint8_t reg1, uint8_t reg2);
int Srr_Mkop_mul (char* buf, uint8_t reg1, uint8_t reg2);
int Srr_Mkop_div (char* buf, uint8_t reg1, uint8_t reg2);
int Srr_Mkop_and (char* buf, uint8_t reg1, uint8_t reg2);
int Srr_Mkop_or (char* buf, uint8_t reg1, uint8_t reg2);
int Srr_Mkop_xor (char* buf, uint8_t reg1, uint8_t reg2);
int Srr_Mkop_shl (char* buf, uint8_t reg1, uint8_t reg2);
int Srr_Mkop_shr (char* buf, uint8_t reg1, uint8_t reg2);
int Srr_Mkop_cpyr (char* buf, uint8_t reg1, uint8_t reg2);
int Srr_Mkop_regunary (char* buf, uint8_t opcode, uint8_t reg);
int Srr_Mkop_not (char* buf, uint8_t opcode, uint8_t reg);
int Srr_Mkop_inc (char* buf, uint8_t opcode, uint8_t reg);
int Srr_Mkop_dec (char* buf, uint8_t opcode, uint8_t reg);
int Srr_Mkop_cpy (char* buf, uint8_t opcode, uint8_t buf1, uint32_t off1,
				  uint8_t buf2, uint32_t off2, uint32_t size);
int Srr_Mkop_cpyas (char* buf, uint8_t buf1, uint32_t off1,
					uint8_t buf2, uint32_t off2, uint32_t size);
int Srr_Mkop_cpysa (char* buf, uint8_t buf1, uint32_t off1,
					uint8_t buf2, uint32_t off2, uint32_t size);
int Srr_Mkop_proxy (char* buf, pid_t proxy);
int Srr_Mkop_noarg (char* buf, uint8_t opcode);
int Srr_Mkop_detach (char* buf);
int Srr_Mkop_disable (char* buf);
int Srr_Mkop_enable (char* buf);
int Srr_Mkop_attach (char* buf, uint8_t irq);
int Srr_Mkop_read (char* buf, int bufno, int offset, int rdlen);
int Srr_Mkop_write (char* buf, int bufno, int offset, int wrlen);

extern int SrrBytecode (int immediate, int irqhandle, void* prog, int len);
extern int SrrIrq_Attach (int irq);
extern int SrrIrq_Detach (int irqhandle);
extern int SrrIrq_Disable (int irqhandle);
extern int SrrIrq_Enable (int irqhandle);
extern int SrrIrq_Handled (int irqhandle, int handled);
extern int SrrIrq_Proxy (int irqhandle, pid_t proxy);
extern int SrrIrq_Mapa (int irqhandle, int bufno, int offset, int size);
extern int SrrIrq_Maps (int irqhandle, int bufno, int size);
extern int SrrIrq_Mapp (int irqhandle, int ioport, int size);
extern int SrrIrq_Ldreg (int irqhandle, int immed, int reg, int value);
extern int SrrIrq_Ldrega (int irqhandle, int immed, int reg,
				   int bufno, int offset, int size);
extern int SrrIrq_Ldregs (int irqhandle, int immed, int reg,
				   int bufno, int offset, int size);
extern int SrrIrq_Ldregp (int irqhandle, int immed, int reg,
				   int port, int size);
extern int SrrIrq_Strega (int irqhandle, int immed, int reg,
				   int bufno, int offset, int size);
extern int SrrIrq_Stregs (int irqhandle, int immed, int reg,
				   int bufno, int offset, int size);
extern int SrrIrq_Stregp (int irqhandle, int immed, int reg,
				   int port, int size);
extern int SrrIrq_Add (int irqhandle, int immed, int reg1, int reg2);
extern int SrrIrq_Sub (int irqhandle, int immed, int reg1, int reg2);
extern int SrrIrq_Mul (int irqhandle, int immed, int reg1, int reg2);
extern int SrrIrq_Div (int irqhandle, int immed, int reg1, int reg2);
extern int SrrIrq_And (int irqhandle, int immed, int reg1, int reg2);
extern int SrrIrq_Or (int irqhandle, int immed, int reg1, int reg2);
extern int SrrIrq_Xor (int irqhandle, int immed, int reg1, int reg2);
extern int SrrIrq_Shl (int irqhandle, int immed, int reg1, int reg2);
extern int SrrIrq_Shr (int irqhandle, int immed, int reg1, int reg2);
extern int SrrIrq_Cpyr (int irqhandle, int immed, int reg1, int reg2);
extern int SrrIrq_Not (int irqhandle, int immed, int reg);
extern int SrrIrq_Inc (int irqhandle, int immed, int reg);
extern int SrrIrq_Dec (int irqhandle, int immed, int reg);
extern int SrrIrq_Cpyas (int irqhandle, int immed, int adbuf, int adoff,
				  int stbuf, int stoff, int size);
extern int SrrIrq_Cpysa (int irqhandle, int immed, int stbuf, int stoff,
				  int adbuf, int adoff, int size);

extern int SrrIrq_Read (int irqhandle, int bufno, int offset, int rdlen,
						char* rdbuf);
extern int SrrIrq_Write (int irqhandle, int bufno, int offset, int rdlen,
						 char* rdbuf);

#endif /* !defined(NO_IRQ) && !defined(__KERNEL__) */
	 
#ifdef __cplusplus
}
#endif

#endif

