/*
 * Implement the little bytecode interpreter that will be used to
 * perform simple operations within an interrupt handler.
 */

#include "module_versions.h"
#include <linux/ioport.h>

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
#include <linux/tqueue.h>
#endif
#include <linux/interrupt.h>
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
#include <asm/io.h>
#else
#include <sys/io.h>
#endif

#include "sysdep.h"

#include	"srr_mod.h"
#include	"srr_tt.h"
#include	"srr_util.h"
#include	"srr_sys.h"
//#include	"../inc/srr.h"
//#include	"../inc/srr_if.h"

typedef struct SBC_OpInfo
{
	int			opcode;
	int			bytelength;
	char		*descr;
} SBC_OpInfo;

static int PerformBytecode (SrrTask* task, int handle, char* cmd, int len);
static int StoreBytecode (SrrTask* task, int irqhandle, char* cmd, int len);
static int SrrBC_Attach (SrrTask* task, int irq);

#ifdef NOT_CURRENTLY_USED
/* This table is in the exact order as the enumeration SrrBC_Opcode,
   and it must stay that way.  We index into this table directly without
   a search, using the opcode as the the index. */

static SBC_OpInfo SrrBC_OpInfo[] = 
{
	{ 0, 0, "Unused" },
	{ SBC_MAPA, 9, "mapa <adbuf:8> <offset:32> <size:32>" },
	{ SBC_MAPS, 5, "maps <stbuf:8> <size:32>" },
	{ SBC_MAPP, 2, "mapp <port:16> <size:16>" },
	{ SBC_LDREG, 5, "ldreg <reg:8> <value:32>" },
	{ SBC_LDREGA, 7, "ldrega <reg:8> <adbuf:8> <offset:32> <size:8>" },
	{ SBC_LDREGS, 7, "ldrega <reg:8> <stbuf:8> <offset:32> <size:8>" },
	{ SBC_LDREGP, 3, "ldregp <reg:8> <port_reg:8> <size:8>" },
	{ SBC_STREGA, 10, "strega <reg:8> <adbuf:8> <offset:32> <size:32>" },
	{ SBC_STREGS, 10, "strega <reg:8> <stbuf:8> <offset:32> <size:32>" },
	{ SBC_STREGP, 3, "stregp <reg:8> <port_reg:8> <size:8>" },
	{ SBC_ADD, 2, "add <rega:8> <regb:8>" },
	{ SBC_SUB, 2, "sub <rega:8> <regb:8>" },
	{ SBC_MUL, 2, "mul <rega:8> <regb:8>" },
	{ SBC_DIV, 2, "div <rega:8> <regb:8>" },
	{ SBC_NOT, 1, "not <reg:8>" },
	{ SBC_AND, 2, "and <rega:8> <regb:8>" },
	{ SBC_OR, 2, "or <rega:8> <regb:8>" },
	{ SBC_SHL, 2, "shl <rega:8> <regb:8>" },
	{ SBC_SHR, 2, "shr <rega:8> <regb:8>" },
	{ SBC_CPYR, 2, "cpyr <rega:8> <regb:8>" },
	{ SBC_CPYSA, 14, "cpyas <adbuf:8> <adoff:32> <stbuf:8> <stoff:32> <size:32>" },
	{ SBC_CPYAS, 14, "cpysa <stbuf:8> <stoff:32> <adbuf:8> <adoff:32> <size:32>" },
	{ SBC_READ, 9, "read <stbuf:8> <stoff:32> <size:32>" },
	{ SBC_WRITE, 9, "write <stbuf:8> <stoff:32> <size:32>" },
	{ SBC_INC, 1, "inc <reg:8>" },
	{ SBC_DEC, 1, "dec <reg:8>" },
	{ SBC_PROXY, 4, "proxy <proxy:32>" },
	{ SBC_ATTACH, 1, "attach <irq:8>" },
	{ SBC_DETACH, 0, "detach" },
	{ SBC_DISABLE, 0, "disable" },
	{ SBC_ENABLE, 0, "enable" },
	{ SBC_XOR, 2, "xor <rega:8> <regb:8>" },
	{ SBC_HANDLED, 1, "handled <bool:8>" },
};
#endif

int Srr_IRQUsed[16] = { 0 };

int SrrIrq_DoIRQCommand (SrrTm* map, SrrTask* task)
{
	int		retval = 0, cmd, immed;

	cmd = task->args.irq.buffer[0];
	immed = (task->args.irq.type == SRRIRQ_IMMEDIATE);

	// printk("SRRIPC: Do IRQ command\n");
	// printk("  Opcode: %d, Length: %d, Immediate: %s\n",
	// cmd, task->args.irq.len, immed ? "yes" : "no");

	if (immed)
	{
		retval = PerformBytecode (task, task->args.irq.irq_handle,
								  task->args.irq.buffer, task->args.irq.len);
	}
	else /* Perform within IRQ handler */
	{
		retval = StoreBytecode (task, task->args.irq.irq_handle,
								task->args.irq.buffer, task->args.irq.len);
	}
	
	return (retval);
}

static int StoreBytecode (SrrTask* task, int irqhandle, char* cmd, int len)
{
	int			retval=0;
	SrrIRQ		*pirq=NULL;
	SrrBC_Program	*prog;
	uint8_t		*line;

	if (task &&
		(irqhandle >= 0 &&
		 irqhandle <= (sizeof(task->IRQArray)/sizeof(SrrIRQ*))))
		pirq = task->IRQArray[irqhandle];

	if (!pirq && cmd[0] != SBC_ATTACH)
		return (-ENOENT);

	prog = &pirq->program;
	
	if (prog->ncode == prog->maxcode)
	{
		uint8_t		**newcode, **oldcode;
		uint32_t	*newcodelen, *oldcodelen;
		
		prog->maxcode += 32;
		newcode = (uint8_t**)xxmalloc(prog->maxcode * sizeof(uint8_t*));
		if (!newcode)
			return (-ENOMEM);
		newcodelen = (uint32_t*)xxmalloc(prog->maxcode * sizeof(uint32_t));
		if (!newcodelen)
		{
			xxfree (newcode);
			return (-ENOMEM);
		}
		
		/* Never leave prog->code in an indeterminate state */
		memcpy (newcode, prog->code, prog->ncode * sizeof(uint8_t*));
		memcpy (newcodelen, prog->codelen, prog->ncode * sizeof(uint32_t));
		oldcode = prog->code;
		oldcodelen = prog->codelen;
		
		prog->code = newcode;
		prog->codelen = newcodelen;
		xxfree (oldcode);
		xxfree (oldcodelen);
	}

	line = (uint8_t*) xxmalloc (len);
	memcpy (line, cmd, len);
	prog->code[prog->ncode++] = line;

	return (retval);
}

static int PerformBytecodeByIRQ (SrrTask* task, SrrIRQ *pirq,
								 char* cmd, int len)
{
	int			retval=0, reg1, reg2;
	int			irqhandle;

	switch (cmd[0])
	{
	  case SBC_MAPA:
	    {
			int					adbuf, i=1, slot;
			uint32_t			offset, nbytes;
			struct resource*	res;

			for (slot=0; slot<SBC_NRES; slot++)
				if (pirq->program.resource[slot] == NULL)
					break;
			if (slot == SBC_NRES)
				retval = -ENOMEM;
			else
			{
				i += SBGET8(adbuf, cmd, i);
				i += SBGET32(offset, cmd, i);
				i += SBGET32(nbytes, cmd, i);
				PDEBUG("Map address %d offset 0x%x size 0x%x\n",
					   adbuf, offset, nbytes);
				if ((res = request_mem_region
					 (offset, nbytes, "srripc")) == NULL)
					retval = -EBUSY;
				else
				{
					pirq->program.resource[slot] = res;
					pirq->program.addr[adbuf].offset = offset;
					pirq->program.addr[adbuf].size = nbytes;
					pirq->program.addr[adbuf].buf = (char*)offset;
				}
			}
			break;
		}
		break;
	  case SBC_MAPP:
	    {
			int			i=1, slot;
			uint32_t	offset, nbytes;
			struct resource*	res;

			for (slot=0; slot<SBC_NRES; slot++)
				if (pirq->program.resource[slot] == NULL)
					break;
			if (slot == SBC_NRES)
				retval = -ENOMEM;
			else
			{
				i += SBGET16(offset, cmd, i);
				i += SBGET16(nbytes, cmd, i);
				PDEBUG("Map ports offset 0x%x size 0x%x\n",
					   offset, nbytes);
				if ((res = request_region (offset, nbytes, "srripc")) == NULL)
					retval = -EBUSY;
				else
					pirq->program.resource[slot] = res;
			}
		}
		break;
	  case SBC_MAPS:
	    {
			int			stno, i=1;
			uint32_t	size;

			i += SBGET8(stno,cmd,i);
			i += SBGET32(size,cmd,i);
			PDEBUG("Map storage %d size 0x%x\n", stno, size);
			
			if (VALIDSTNO(stno))
			{
				if (pirq->program.stor[stno].buf)
				{
					xxfree (pirq->program.stor[stno].buf);
					pirq->program.stor[stno].buf = NULL;
					pirq->program.stor[stno].size = 0;
				}
				pirq->program.stor[stno].buf = xxmalloc (size);

				// printk ("SRRIRQ: MAPS %d %d\n", stno, size);
				
				if (!pirq->program.stor[stno].buf)
					retval = -ENOMEM;
				else
				{
					pirq->program.stor[stno].size = size;
					memset (pirq->program.stor[stno].buf, 0, size);
				}
			}
			else
				retval = -EINVAL;
		}
		break;
	  case SBC_LDREG:
	    {
			int			i=1;
			int32_t		value;

			i += SBGET8(reg1,cmd,i);
			i += SBGET32(value,cmd,i);
			PDEBUG("ldreg %d %d\n", reg1, value);

			if (VALIDREG(reg1))
				pirq->program.reg[reg1] = value;
			else
				retval = -EINVAL;
		}
		break;
	  case SBC_LDREGA:
	  case SBC_LDREGS:
	    {
			int			stno, size, i=1, valid=0;
			uint32_t	offset;
			char		*buf;

			i += SBGET8(reg1,cmd,i);
			i += SBGET8(stno,cmd,i);
			i += SBGET32(offset,cmd,i);
			i += SBGET8(size,cmd,i);
			PDEBUG("ldreg%s %d %d %d %d\n",
				   cmd[0] == SBC_LDREGA ? "a" : "s",
				   reg1, stno, offset, size);

			if (cmd[0] == SBC_LDREGA)
				valid = (VALIDADDR(pirq,stno) &&
						 (offset + size <= pirq->program.addr[stno].size));
			else
				valid = (VALIDSTOR(pirq,stno) &&
						 (offset + size <= pirq->program.stor[stno].size));
			
			if (valid && VALIDREG(reg1))
			{
				if (cmd[0] == SBC_LDREGA)
					buf = pirq->program.addr[stno].buf;
				else
					buf = pirq->program.stor[stno].buf;

				switch (size)
				{
				  case sizeof(uint8_t):
					SBGET8(pirq->program.reg[reg1], buf, offset);
					break;
				  case sizeof(uint16_t):
					SBGET16(pirq->program.reg[reg1], buf, offset);
					break;
				  case sizeof(uint32_t):
					SBGET32(pirq->program.reg[reg1], buf, offset);
					break;
				  default:
					retval = -EINVAL;
					break;
				}
			}
			else
				retval = -EINVAL;
		}
		break;
	  case SBC_LDREGP:
	    {
			int				i=1;
			unsigned int	port, size;

			i += SBGET8(reg1, cmd, i);
			i += SBGET8(reg2, cmd, i);
			i += SBGET8(size, cmd, i);
			PDEBUG("ldregp %d %d %d\n", reg1, reg2, size);

			port = pirq->program.reg[reg2];
			
			if (VALIDREG(reg1))
			{
				switch (size)
				{
				  case sizeof(uint8_t):
					pirq->program.reg[reg1] = inb (port);
					break;
				  case sizeof(uint16_t):
					pirq->program.reg[reg1] = inw (port);
					break;
				  case sizeof(uint32_t):
					pirq->program.reg[reg1] = inl (port);
					break;
				  default:
					retval = -EINVAL;
				}
			}
			else
				retval = -EINVAL;
		}
		break;
	  case SBC_STREGA:
	  case SBC_STREGS:
	    {
			int			stno, size, i=1, valid=0;
			uint32_t	offset;
			char		*buf;

			i += SBGET8(reg1,cmd,i);
			i += SBGET8(stno,cmd,i);
			i += SBGET32(offset,cmd,i);
			i += SBGET8(size,cmd,i);
			PDEBUG("streg%s %d %d %d %d\n",
				   cmd[0] == SBC_STREGA ? "a" : "s",
				   reg1, stno, offset, size);

			if (cmd[0] == SBC_STREGA)
				valid = (VALIDADDR(pirq,stno) &&
						 (offset + size <= pirq->program.addr[stno].size));
			else
				valid = (VALIDSTOR(pirq,stno) &&
						 (offset + size <= pirq->program.stor[stno].size));
			
			if (valid && VALIDREG(reg1))
			{
				if (cmd[0] == SBC_STREGA)
					buf = pirq->program.addr[stno].buf;
				else
					buf = pirq->program.stor[stno].buf;

				switch (size)
				{
				  case sizeof(uint8_t):
					SBPUT8(pirq->program.reg[reg1], buf, offset);
					break;
				  case sizeof(uint16_t):
					SBPUT16(pirq->program.reg[reg1], buf, offset);
					break;
				  case sizeof(uint32_t):
					SBPUT32(pirq->program.reg[reg1], buf, offset);
					break;
				  default:
					retval = -EINVAL;
					break;
				}
			}
			else
				retval = -EINVAL;
		}
		break;
	  case SBC_STREGP:
	    {
			int				i=1;
			unsigned int	port, size;

			i += SBGET8(reg1, cmd, i);
			i += SBGET8(reg2, cmd, i);
			i += SBGET8(size, cmd, i);
			PDEBUG("stregp %d %d %d\n", reg1, reg2, size);

			port = pirq->program.reg[reg2];
			
			if (VALIDREG(reg1))
			{
				switch (size)
				{
				  case sizeof(uint8_t):
					outb (pirq->program.reg[reg1], port);
					break;
				  case sizeof(uint16_t):
					outw (pirq->program.reg[reg1], port);
					break;
				  case sizeof(uint32_t):
					outl (pirq->program.reg[reg1], port);
					break;
				  default:
					retval = -EINVAL;
				}
			}
			else
				retval = -EINVAL;
		}
		break;
	  case SBC_ADD:
	  case SBC_SUB:
	  case SBC_MUL:
	  case SBC_DIV:
	  case SBC_AND:
	  case SBC_OR:
	  case SBC_XOR:
	  case SBC_SHL:
	  case SBC_SHR:
	  case SBC_CPYR:
	    {
			int			i=1;
			i+=SBGET8(reg1, cmd, i);
			i+=SBGET8(reg2, cmd, i);
			if (VALIDREG(reg1) && VALIDREG(reg2))
			{
				switch (cmd[0])
				{
				  case SBC_ADD:
					PDEBUG("add %d %d\n", reg1, reg2);
					pirq->program.reg[reg1] += pirq->program.reg[reg2];
					break;
				  case SBC_SUB:
					PDEBUG("sub %d %d\n", reg1, reg2);
					pirq->program.reg[reg1] -= pirq->program.reg[reg2];
					break;
				  case SBC_MUL:
					PDEBUG("mul %d %d\n", reg1, reg2);
					pirq->program.reg[reg1] *= pirq->program.reg[reg2];
					break;
				  case SBC_DIV:
					PDEBUG("div %d %d\n", reg1, reg2);
					if (pirq->program.reg[reg2] != 0)
						pirq->program.reg[reg1] /= pirq->program.reg[reg2];
					else
						retval = -ERANGE;
					break;
				  case SBC_AND:
					PDEBUG("and %d %d\n", reg1, reg2);
					pirq->program.reg[reg1] &= pirq->program.reg[reg2];
					break;
				  case SBC_OR:
					PDEBUG("or %d %d\n", reg1, reg2);
					pirq->program.reg[reg1] |= pirq->program.reg[reg2];
					break;
				  case SBC_XOR:
					PDEBUG("xor %d %d\n", reg1, reg2);
					pirq->program.reg[reg1] ^= pirq->program.reg[reg2];
					break;
				  case SBC_SHL:
					PDEBUG("shl %d %d\n", reg1, reg2);
					pirq->program.reg[reg1] <<= pirq->program.reg[reg2];
					break;
				  case SBC_SHR:
					PDEBUG("shr %d %d\n", reg1, reg2);
					pirq->program.reg[reg1] >>= pirq->program.reg[reg2];
					break;
				  case SBC_CPYR:
					PDEBUG("cpyr %d %d\n", reg1, reg2);
					pirq->program.reg[reg1] = pirq->program.reg[reg2];
					break;
				}
			}
			else
			{
				retval = -EINVAL;
			}
		}
		break;
	  case SBC_CPYSA:
	    {
			int			adbuf, stbuf, i=1;
			uint32_t	adoff, stoff, size;

			i += SBGET8(adbuf,cmd,i);
			i += SBGET32(adoff,cmd,i);
			i += SBGET8(stbuf,cmd,i);
			i += SBGET32(stoff,cmd,i);
			i += SBGET32(size,cmd,i);
			
			PDEBUG("cpysa %d %d %d %d %d\n", adbuf, adoff, stbuf, stoff, size);
			
			if (VALIDADDR(pirq,adbuf) && VALIDSTOR(pirq,stbuf) &&
				(stoff + size <= pirq->program.stor[stbuf].size) &&
				(adoff + size <= pirq->program.addr[adbuf].size))
			{
				memcpy (&pirq->program.stor[stbuf].buf[stoff],
						&pirq->program.addr[adbuf].buf[adoff],
						size);
			}
			else
				retval = -EINVAL;
		}
		break;
	  case SBC_CPYAS:
	    {
			int			adbuf, stbuf, i=1;
			uint32_t	adoff, stoff, size;

			i += SBGET8(stbuf,cmd,i);
			i += SBGET32(stoff,cmd,i);
			i += SBGET8(adbuf,cmd,i);
			i += SBGET32(adoff,cmd,i);
			i += SBGET32(size,cmd,i);

			PDEBUG("cpyas %d %d %d %d %d\n", stbuf, stoff, adbuf, adoff, size);
			
			if (VALIDADDR(pirq,adbuf) && VALIDSTOR(pirq,stbuf) &&
				(stoff + size <= pirq->program.stor[stbuf].size) &&
				(adoff + size <= pirq->program.addr[adbuf].size))
			{
				memcpy (&pirq->program.addr[adbuf].buf[adoff],
						&pirq->program.stor[stbuf].buf[stoff],
						size);
			}
			else
				retval = -EINVAL;
		}
		break;
	  case SBC_READ:
	    {
			int			stno, i=1;
			uint32_t	offset, size;
			
			i += SBGET8(stno,cmd,i);
			i += SBGET32(offset,cmd,i);
			i += SBGET32(size,cmd,i);

			PDEBUG("read %d %d %d\n", stno, offset, size);
			
			if (VALIDSTOR(pirq,stno) &&
				(offset + size <= pirq->program.stor[stno].size) &&
				task->args.irq.extra)
			{
				if (size <= task->args.irq.extralen)
				{
					cpy_k2u (task->args.irq.extra,
							 &pirq->program.stor[stno].buf[offset],
							 size);
				}
				else
					retval = -ENOMEM;
			}
			else
				retval = -EINVAL;
		}
		break;
	  case SBC_WRITE:
	    {
			int			stno, i=1;
			uint32_t	offset, size;
			
			i += SBGET8(stno,cmd,i);
			i += SBGET32(offset,cmd,i);
			i += SBGET32(size,cmd,i);

			PDEBUG("write %d %d %d\n", stno, offset, size);
			
			if (VALIDSTOR(pirq,stno) &&
				(offset + size <= pirq->program.stor[stno].size) &&
				task->args.irq.extra)
			{
				if (size <= task->args.irq.extralen)
				{
					cpy_u2k (&pirq->program.stor[stno].buf[offset],
							 task->args.irq.extra,
							 size);
				}
				else
					retval = -ENOMEM;
			}
			else
				retval = -EINVAL;
		}
		break;
		
	  case SBC_INC:
	  case SBC_DEC:
	  case SBC_NOT:
	    {
			SBGET8(reg1, cmd, 1);
			if (VALIDREG(reg1))
			{
				switch (cmd[0])
				{
				  case SBC_INC:
					PDEBUG("inc %d\n", reg1);
					pirq->program.reg[reg1]++;
					break;
				  case SBC_DEC:
					PDEBUG("dec %d\n", reg1);
					pirq->program.reg[reg1]--;
					break;
				  case SBC_NOT:
					PDEBUG("not %d\n", reg1);
					pirq->program.reg[reg1] = ~pirq->program.reg[reg1];
					break;
				}
			}
			else
				retval = -EINVAL;
		}
		break;
	  case SBC_PROXY:
	    {
			int		proxy;
			SBGET32 (proxy, cmd, 1);
			PDEBUG("proxy %d\n", proxy);
			if (pirq)
				pirq->proxy = proxy;
			else
				retval = -ENOENT;
		}  
		break;
	  case SBC_ATTACH:
	    {
			int				irq;
			int				handle;
			
			SBGET8(irq, cmd, 1);
			PDEBUG("attach %d\n", irq);
			handle = SrrBC_Attach (task, irq);
			if (handle < 0)
				retval = handle;
			else
				task->args.irq.irq_handle = handle;
		}
		break;
	  case SBC_DETACH:
		if (task)
		{
			irqhandle = task->args.irq.irq_handle;
			PDEBUG("detach %d\n", irqhandle);
			retval = SrrIrq_Detach (task, irqhandle);
		}
		else
		{
			retval = -EINVAL;
		}
		break;
	  case SBC_DISABLE:
		pirq->flags |= SBC_FLAG_DISABLED;
		break;
	  case SBC_ENABLE:
		pirq->flags &= ~SBC_FLAG_DISABLED;
		break;
	  case SBC_HANDLED:
	    {
			int			handled;
			
			SBGET8(handled,cmd,1);
			pirq->flags &= ~SBC_FLAG_HANDLED;
			pirq->flags |= (handled ? SBC_FLAG_HANDLED : 0);
		}
		break;
	}
	/* We are not allowed to return EBUSY.  It is reserved by
	   the library to determine that a child process has tried to
	   write to its parent's file descriptor.  It should have been
	   EACCESS in the module, which would have made more sense. */

	if (retval == -EBUSY)
		retval = -EACCES;

	return (retval);
}

static int SrrIrq_RunIRQCode (SrrIRQ *pirq)
{
	int			i, retval=0;
	
	// printk ("Run %d lines of code\n", pirq->program.ncode);
	for (i=0; i<pirq->program.ncode; i++)
	{
		retval = PerformBytecodeByIRQ (NULL, pirq,
									   pirq->program.code[i],
									   pirq->program.codelen[i]);
	}
	return (0);
}

static int PerformBytecode (SrrTask* task, int irqhandle, char* cmd, int len)
{
	int			retval=0;
	SrrIRQ		*pirq=NULL;

	if (task &&
		(irqhandle >= 0 &&
		 irqhandle <= (sizeof(task->IRQArray)/sizeof(SrrIRQ*))))
		pirq = task->IRQArray[irqhandle];

	if (!pirq && cmd[0] != SBC_ATTACH)
		return (-ENOENT);

	retval = PerformBytecodeByIRQ (task, pirq, cmd, len);
	return (retval);
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
#define IRQTYPE	irqreturn_t
#define IRQRETURN return ((pirq->flags & SBC_FLAG_HANDLED) ? IRQ_HANDLED : IRQ_NONE)
#else
#define IRQTYPE void
#define IRQRETURN
#endif

static IRQTYPE SrrIrq_Handler (int irq, void* virq
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,20)
							   , struct pt_regs *regs_are_unused
#endif
	)
{
	SrrIRQ		**ppirq = (SrrIRQ**) virq;
	SrrIRQ		*pirq = *ppirq;

	if (!pirq || (pirq->flags & SBC_FLAG_DISABLED))
	{
		// Do nothing
	}
	else
	{
		atomic_inc(&pirq->count);
		atomic_inc(&pirq->pending);
		SrrIrq_RunIRQCode (pirq);
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
		schedule_task (&pirq->bhtask);
#else
		schedule_work (&pirq->bhwork);
#endif
	}

	IRQRETURN;
}

// This function is called from a work_queue
#ifdef WORK_DATA_IS_VOID
static void SrrIrq_BottomHalfHandler (void* data)
#else
static void SrrIrq_BottomHalfHandler (struct work_struct *work)
#endif // WORK_DATA_IS_VOID
{
#ifdef WORK_DATA_IS_VOID
	SrrIRQ		**ppirq = (SrrIRQ**) data;
	SrrIRQ		*pirq = *ppirq;
#else
	SrrIRQ		*pirq = container_of (work, SrrIRQ, bhwork);
#endif // WORK_DATA_IS_VOID
	int			err;
	int			pending;

	extern SrrTm* srr_taskmap;

	if (in_interrupt())
	{
		Oops ("IRQ bottom half handler called in interrupt context\n");
	}
	if (pirq == NULL)
		return;

	atomic_inc (&pirq->bhcount);
	pending = atomic_read(&pirq->pending);
	
	// printk ("IRQ %d\n", pirq->irq);
	if (pirq->proxy != 0)
	{
		int		i;
		
		/* Trigger the proxy once per interrupt. */
		for (i=0; i<pending; i++)
		{
			atomic_dec (&pirq->pending);
			err = SrrDoTrigger(srr_taskmap, pirq->proxy, 0);
			if (err != 0)
			{
				Oops ("Interrupt %d trigger proxy %d returned error %d\n",
					  pirq->irq, pirq->proxy, err);
			}
		}
	}
}

// Older kernels use SA_SHIRQ instead of IRQF_SHARED
#ifndef IRQF_SHARED
#define IRQF_SHARED SA_SHIRQ
#endif

static int SrrBC_Attach (SrrTask* task, int irq)
{
	SrrIRQ		*pirq;
	int			handle, maxhandle, retval;
	
	// printk ("SRRIRQ: Attach %d\n", irq);

	maxhandle = sizeof (task->IRQArray) / sizeof(SrrIRQ*);
	for (handle=0; handle<maxhandle; handle++)
		if (task->IRQArray[handle] == NULL)
			break;

	if (handle == maxhandle)
		return (-ENOMEM);

	if (irq < -1 || irq > 15)
		return (-EINVAL);
	
	pirq = xxmalloc (sizeof(SrrIRQ));
	memset (pirq, 0, sizeof(SrrIRQ));

	/* initial flag set: default to marking this irq handler as the
       definitive irq handler (the one that tells the OS that the irq
       is handled) */
	pirq->flags = SBC_FLAG_HANDLED;
	pirq->irq = irq;
	task->IRQArray[handle] = pirq;

	if (irq >= 0 && irq <= 15)
	{
		// printk ("SRRIRQ: request IRQ %d\n", irq);
		retval = request_irq (irq, SrrIrq_Handler, IRQF_SHARED, "srripc",
							  &task->IRQArray[handle]);
		if (retval == 0)
		{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
			pirq->bhtask.routine = SrrIrq_BottomHalfHandler;
			/* We pass the pointer to the location that the IRQ pointer is
			   held in, which means that we can check for the presence of
			   the structure during the IRQ handler. */
			pirq->bhtask.data = &task->IRQArray[handle];
#else
			/* We pass the pointer to the location that the IRQ pointer is
			   held in, which means that we can check for the presence of
			   the structure during the IRQ handler. */
#ifdef WORK_DATA_IS_VOID
			INIT_WORK (&pirq->bhwork, SrrIrq_BottomHalfHandler,
					   &task->IRQArray[handle]);
#else
			INIT_WORK (&pirq->bhwork, SrrIrq_BottomHalfHandler);
#endif // WORK_DATA_IS_VOID
#endif
			Srr_IRQUsed[irq] ++;
		}
		else
		{
			xxfree (pirq);
			task->IRQArray[handle] = NULL;
			handle = retval;
			// printk ("SRRIRQ: failed request: %d\n", retval);
		}
	}
	
	return (handle);
}

void SrrIrq_DetachAll (SrrTask* task)
{
	int		i, mxi;

	mxi = sizeof(task->IRQArray) / sizeof(SrrIRQ*);
	for (i=0; i<mxi; i++)
	{
		SrrIrq_Detach (task, i);
	}
}

int SrrIrq_Detach (SrrTask* task, int handle)
{
	SrrIRQ		*pirq;
	int			maxhandle, i, irq;
	
	maxhandle = sizeof (task->IRQArray) / sizeof(SrrIRQ*);
	if (handle >= maxhandle || handle < 0)
		return (-EINVAL);
	if (task->IRQArray[handle] == NULL)
		return (-ENOENT);

	pirq = task->IRQArray[handle];
	irq = pirq->irq;
	if (irq >= 0 && irq <= 15)
	{
		// printk ("SRRIRQ: Detach %d\n", handle);
		disable_irq(irq);
		free_irq (pirq->irq, &task->IRQArray[handle]);
		task->IRQArray[handle] = 0;
		enable_irq(irq);
		// printk ("free_irq %d\n", pirq->irq);

		/* How do we ensure that the interrupt is not called while we
		   are busy detaching it?  Do we need to care?  Yes.  The bottom
		   half handler could run after we zero the interrupt. */
		
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
		flush_scheduled_tasks();
#else
		flush_scheduled_work();
#endif

		Srr_IRQUsed[pirq->irq] --;
	}

	if (pirq->program.code)
	{
		for (i=0; i<pirq->program.ncode; i++)
			xxfree (pirq->program.code[i]);
		xxfree (pirq->program.code);
	}

	for (i=0; i<SBC_NRES; i++)
	{
		if (pirq->program.resource[i])
		{
			release_resource (pirq->program.resource[i]);
			// pirq->program.resource[i] = NULL;
		}
	}

	for (i=0; i<SBC_NSTOR; i++)
	{
		if (pirq->program.stor[i].buf)
		{
			xxfree (pirq->program.stor[i].buf);
			// pirq->program.stor[i].buf = NULL;
			// pirq->program.stor[i].size = 0;
		}
	}

	for (i=0; i<SBC_NADDR; i++)
	{
		/* These are freed by release_resource above */
		// pirq->program.addr[adbuf].offset = 0;
		// pirq->program.addr[adbuf].size = 0;
		// pirq->program.addr[adbuf].buf = NULL;
	}
	
	xxfree (pirq);
	task->IRQArray[handle] = 0;
	return (0);
}
