/*
 * srr_tt.h: interface to the task structures
 *
 *  Linux Loadable Module Implementation of QNX Send/Receive/Reply
 *	message passing, and related function.
 *
 *	QNX is a trademark of QSSL, Ltd. <www.qnx.com>
 *
 * Copyright (C) 2000 Cogent Real-Time Systems Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software in a file called 'COPYING'; if not, write to:
 * 	Free Software Foundation, Inc.,
 * 	59 Temple Place, Suite 330,
 * 	Boston, MA 
 * 	02111-1307  USA
 *
 * $Id: srr_tt.h 2870 2006-07-19 12:03:10Z andrew $
 */

#ifndef SRR_TT_H
#define SRR_TT_H

#include <linux/types.h>
#include "srr_taskLocks.h"

/*
 * kmalloc priority level - currently atomic, because I don't want to have
 * to think about what will happen if we get rescheduled at any point
 * where kmalloc() could be called from.
 */

#define xxmalloc(S)	kmalloc(S, GFP_ATOMIC)
#define xxfree(A)	kfree(A)

#ifndef NO_IRQ
#include "srr_bytecode.h"
#include "srr_irq.h"
#endif

#include "srr_if.h"

/*
 * Older kernels need an IRQ-approved version of queue_task
 */

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,4,0)
#define queue_task_irq(a,b)		queue_task(a,b)
#endif

/*
 * SrrBuf: buffers data in the kernel that is being xferred between tasks.
 */

typedef struct SrrBuf
{
	void*	data;
	size_t	datalen;
	size_t	maxsize;
} SrrBuf;

/* 2000/03/26 - Alessandro Sala */
int 	SrrBuf_AdjustSize(SrrBuf* buf, size_t len);
int 	SrrBuf_OffsetRead(SrrBuf* buf, size_t offset, char* from, size_t size);
int	SrrBuf_OffsetWrite(SrrBuf* buf, size_t offset, char* from, size_t size);

int		SrrBuf_Read(SrrBuf* buf, char* from, size_t size);
int		SrrBuf_Write(SrrBuf* buf, char* from, size_t size);
int		SrrBuf_Release(SrrBuf* buf);
int		SrrBuf_LocalRead(SrrBuf* buf, char* from, size_t size);
int		SrrBuf_LocalWrite(SrrBuf* buf, char* to, size_t size);
void	SrrBuf_Reset (SrrBuf* buf);

typedef int (*pfnSrrBufRead)(SrrBuf*, char*, size_t);
typedef int (*pfnSrrBufWrite)(SrrBuf*, char*, size_t);

/*
 * SrrRq: routines to manipulate a recv_q
 */
struct SrrTask**	SrrRq_Find(struct SrrTask* server, pid_t pid);
struct SrrTask*		SrrRq_Unlink(struct SrrTask** taskp);
struct SrrTask*		SrrRq_UnlinkNdebug(struct SrrTask** taskp);
void				SrrRq_Queue (struct SrrTask* server, struct SrrTask* task);
void				SrrRq_QueueNdebug (struct SrrTask* server,
									   struct SrrTask* task);
struct SrrTask** 	SrrRq_First (struct SrrTask* server);
int SrrRq_NonEmpty (struct SrrTask* server);

/*
 * SrrTm: a pid to task map.
 */
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,3,0)
#define UNINITIALIZED_WAIT_QUEUE(x)	wait_queue_head_t x
#else
#define UNINITIALIZED_WAIT_QUEUE(x)	struct wait_queue *x
#endif

#ifdef TASK_SPECIFIC_LOCKS
struct SrrDeleteQueue
{
  struct DeleteNode *head;
  struct DeleteNode *tail;
  struct mutex       lock;
};

struct DeleteNode
{
  struct DeleteNode *next;
  struct DeleteNode *parent;
  struct SrrTask *task;
};
#endif

typedef struct SrrTm
{
	struct SrrTask** 	tasks;
#ifdef TASK_SPECIFIC_LOCKS
  struct SrrDeleteQueue *deleteQueue;
  UNINITIALIZED_WAIT_QUEUE(Map_wq);
  atomic_t map_readers;
  atomic_t map_writers;
#endif
} SrrTm;

#ifdef TASK_SPECIFIC_LOCKS
typedef struct SrrMutex
{
#ifdef LOCK_VIA_SPINLOCKS
  spinlock_t lock;
  unsigned long flags;
#else
	struct mutex mutex;
#endif /* LOCK_VIA_SPINLOCKS */
	atomic_t	 users;
  UNINITIALIZED_WAIT_QUEUE(wq);
} SrrMutex;
#endif /* TASK_SPECIFIC_LOCKS */

SrrTm*			SrrTm_new(void);
void			SrrTm_delete(SrrTm* map);
struct SrrTask*	SrrTm_Get(SrrTm* map, pid_t pid);
struct SrrTask*	SrrTm_ProxyAlloc(SrrTm* map);
struct SrrTask*	SrrTm_Alloc(SrrTm* map, pid_t pid, int proxy);
int				SrrTm_Free(SrrTm* map, pid_t pid);
int				SrrTm_Invariant(SrrTm* map);
int				SrrTm_Kill(SrrTm* map, pid_t pid, int signo);

/*
 * Task: holds state for a Linux process participating in SRR.
 */

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,3,0)
#define UNINITIALIZED_WAIT_QUEUE(x)	wait_queue_head_t x
#else
#define UNINITIALIZED_WAIT_QUEUE(x)	struct wait_queue *x
#endif

typedef struct SrrTask
{
	struct SrrTask		*target; 	// task we're a proxy to
#define IS_PROXY(T) (T->target ? 1 : 0)
#ifdef TASK_SPECIFIC_LOCKS
	struct SrrMutex		mutex;
#endif
	struct timer_list	timer;		// timer for this proxy
	unsigned long		repeat;		// repeat time in jiffies
	atomic_t		count;		// proxy Trigger() count
	int			firecount;	// total number of times proxy has fired

	pid_t	pid;

	pid_t	waker;	// pid of the task that called wake_up(), for info only */

	UNINITIALIZED_WAIT_QUEUE(wq);
	UNINITIALIZED_WAIT_QUEUE(poll_q);

	int	last_ioctl;
	
	union {
		pid_t	pid;

		struct SrrTraceArgs	trace;
		struct SrrKillArgs	kill;

		struct SrrSendArgs	send;
		struct SrrRecvArgs	recv;
		struct SrrReplyArgs	reply;
		/* 2005/09/05 - Len Meakin */
		struct SrrRelayArgs relay;
		/* 2000/02/17 - Alessandro Sala */
		struct SrrReadmsgArgs	readmsg;
		/* 2000/03/27 - Alessandro Sala */
		struct SrrWritemsgArgs	writemsg;

		/* 2000/10/15 - John Green */
		struct SrrSendmxArgs		sendmx;
		struct SrrRecvmxArgs		recvmx;
		struct SrrReplymxArgs		replymx;
		struct SrrReadmsgmxArgs		readmsgmx;
		struct SrrWritemsgmxArgs	writemsgmx;

		struct SrrNameAttachArgs	name_a;
		struct SrrNameLocateArgs	name_l;
		struct SrrNameDetachArgs	name_d;
		
		struct SrrProxyAttachArgs	proxy_a;
		struct SrrProxyDetachArgs	proxy_d;
		struct SrrTriggerArgs		trigger;
		struct SrrPFlagsArgs		pflags;
		struct SrrTimerCreateArgs	tcreate;
		struct SrrTimerDeleteArgs	tdelete;
		struct SrrTimerSetArgs		tset;
		struct SrrTimerGetArgs		tget;
		struct SrrClockGetresArgs	cgetres;
		struct SrrIRQArgs			irq;
		struct SrrRecvTimeoutArgs	recvTimeout;
    struct SrrSendTimeoutArgs sendTimeout;
	  struct SrrSendmxTimeoutArgs sendmxTimeout;
    struct SrrRecvmxTimeoutArgs recvmxTimeout;
  } args;
	
	/* the fact that values > 0 are blocked is used */
	enum {
		RELEASED = -1,	/* released due to task death (an ESRCH) */
		READY = 0,		/* task not waiting */
		SEND_BLOCKED,	/* task called Send(), waiting for a Receive() */
		RECV_BLOCKED,	/* task called Receive(), waiting for a Send() */
		REPLY_BLOCKED,	/* task called Send(), waiting for a Reply() */
		RECV_SPINLOCKED, /* task is in a spinlock for Receive() */
		TRIGGER_SPINLOCKED, /* task is in a spinlock for Trigger() */
	} state;

	unsigned long	flags;	/* flags, as in enum_FLAGS */
	
	struct SrrTask*	recv_q;	/* queue of tasks send-blocked on us */
	struct SrrTask*	recv_l;	/* link in recv_q of task we're send-blocked on */

	SrrBuf	buf;
#ifndef NO_IRQ
	SrrIRQ	*IRQArray[8];
#endif
	struct task_struct *kern_task;
} SrrTask;

const char*	SrrTaskState2Str(SrrTask* task, int state);

SrrTask*	SrrTask_new(pid_t pid);
void		SrrTask_delete(SrrTask* task, SrrTm *map, int recurse);

/*
 * SrrNm(name map): underlying attach/detach/locate implementation
 */

/* this could be dynamically set at module load time instead */
#define SRR_MAX_NAMES	256

struct SrrNameIdMap
{
	SrrTask*	task;
	char*		name;
};

extern struct SrrNameIdMap srr_NameIdMap[SRR_MAX_NAMES];

int		SrrNm_Attach(SrrTask* task, char* name);
int		SrrNm_Locate(SrrTm* map, const char* name);
int		SrrNm_Detach(SrrTask* task, int name_id);
void	SrrNm_Free(pid_t pid);

#endif

#ifdef TASK_SPECIFIC_LOCKS
void addToDeleteQueue(SrrTask *task, SrrTm *map);
int cleanDeleteQueue(SrrTm *map);
struct SrrTask*	SrrTm_GetLocked(SrrTm* map, pid_t pid);
#else
struct SrrTask*	SrrTm_GetLocked(SrrTm* map, pid_t pid);
#endif
#ifdef PRINT_TASK_TABLE
void printTm(SrrTm *map);
#endif
