/*
 * srr_util.c:
 *
 *  Linux Loadable Module Implementation of QNX Send/Receive/Reply
 *	message passing, and related function.
 *
 *	QNX is a trademark of QSSL, Ltd. <www.qnx.com>
 *
 * Copyright (C) 2000 Cogent Real-Time Systems Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software in a file called 'COPYING'; if not, write to:
 * 	Free Software Foundation, Inc.,
 * 	59 Temple Place, Suite 330,
 * 	Boston, MA 
 * 	02111-1307  USA
 *
 * $Id: srr_util.c 222 2011-08-19 13:41:03Z meaton $
 */

#include "module_versions.h"
#include "srr_util.h"
#include <linux/pagemap.h>
#include <linux/mm.h>
#include <linux/vmalloc.h> 

/* HEXDUMP */
#define BYTES_PER_LINE 16
#define BYTES_PER_DIVISION 8
#define HEX_STR_LEN 64

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
/* Remove any previous definitions.  */
#undef	alloca
extern void *alloca (size_t __size);
#ifdef	__GNUC__
# define alloca(size)	__builtin_alloca (size)
#endif /* GCC.  */
#else
#include <alloca.h>
#endif /* Kernel version 2.6.0 onward. */


/*
 * In the 2.6 kernels, this function is provided but not exported, so
 * kernel modules cannot use it.  So, we copy it here.
 */

static int srr_copy_from_user (void* dst, const void* src, size_t size)
{
	return (copy_from_user (dst, src, size));
}

static int srr_copy_to_user (void* dst, const void* src, size_t size)
{
   return (copy_to_user (dst, src, size));
}

static int srr_safe_copy (void* dest, const void* src, size_t length,
						  int touser,
						  int (*pfcopy)(void*,const void*,size_t))
{
	int						ret=0, npages, result=0;
	struct page				**pages;
	struct vm_area_struct	**vmas;
	unsigned long			addr;
	unsigned long			end;
	int						usedvmalloc = 0;
  
  addr = (touser ? (unsigned long)dest : (unsigned long)src);
	end = addr + length;

	if (addr >= end)
	{
		Oops ("Bad srr_safe_copy call: start(%lu) > end (%lu)\n",
			  addr, end);
		result = -EINVAL;
	}
	else
	{
		npages = (end+PAGE_SIZE-1)/PAGE_SIZE-addr/PAGE_SIZE;

		// use vmalloc instead of alloca for bigger messages
		if (npages < 100)
		{
			PDEBUG ("srr_safe_copy - using alloca to allocate %d bytes for tables with %d pageentries ...\n", npages*(sizeof(struct page*)+sizeof(struct vm_area_struct*)), npages);
			pages = (struct page**) alloca(npages * sizeof(struct page*));
			vmas = (struct vm_area_struct**) alloca(npages * sizeof(struct vm_area_struct*));
		}
		else
		{
			PDEBUG ("srr_safe_copy - using vmalloc to allocate %d bytes for tables with %d pageentries ...\n", npages*(sizeof(struct page*)+sizeof(struct vm_area_struct*)), npages);
			pages = (struct page**) vmalloc(npages * sizeof(struct page*));
			vmas = (struct vm_area_struct**) vmalloc(npages * sizeof(struct vm_area_struct*));
			usedvmalloc=1;
		}

		// Nov 13/08 - Charlie Brady
		// protect get_user_pages by taking a read-only lock on
		// the vmem semaphore.  Release the semaphore after the
		// copy (could the memory be mapped out in the meantime
		// otherwise?) and release before calling Oops, though we
		// have no evidence that it might matter.
#undef USE_MEM_SEMAPHORE
#ifdef USE_MEM_SEMAPHORE
		down_read(&current->mm->mmap_sem);
#endif // USE_MEM_SEMAPHORE
    ret = get_user_pages(current, current->mm, addr,
							 npages, touser ? 1 : 0, 1, pages, vmas);
		
		if (ret == npages)
		{
			result = pfcopy (dest, src, length);
			result = 0;
#ifdef USE_MEM_SEMAPHORE
			up_read(&current->mm->mmap_sem);
#endif // USE_MEM_SEMAPHORE
		}
		else // mapped the wrong number of pages, or failed
		{
#ifdef USE_MEM_SEMAPHORE
			up_read(&current->mm->mmap_sem);
#endif // USE_MEM_SEMAPHORE
			if (ret >= 0)
			{
				Oops ("get_user_pages: expect %d pages, mapped %d pages\n",
					  npages, ret);
				result = -ENOMEM;
			}
			else if (ret == -512)
			{
				PDEBUG("get_user_pages for/on a task with a fatal signal, %d\n", ret);
				result = ret;
			}
			else
			{
				Oops ("get_user_pages: error %d\n", ret);
				result = ret;
			}
		}

		while (ret-- > 0)
		{
			page_cache_release (pages[ret]);
		}

		// dont forget to free ...
		if (usedvmalloc)
		{
			vfree(pages);
			vfree(vmas);
		}
	}
  
  return (result);
}

#ifdef SRR_DEBUG
// was 2, but that is too noisy.  Use exe/srrctl -t 2 to change this.
int srr_debug_level = 0;
#else
int srr_debug_level = 0;
#endif

#ifdef SRR_DEBUG
int srr_debug_trace(int level)
{
	int last = srr_debug_level;

	srr_debug_level = level;

	return last;
}
#endif

/*
 * memory xfer: wrappers for copying data between user and kernel space
 */

int cpy_u2k(void* dst, const void* src, size_t size)
{
	
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)
	int		result = 0;
  if (size > 0)
	{
		if ((result = srr_safe_copy (dst, src, size, 0,
									 srr_copy_from_user)) != 0)
		{
			if (result < 0)
         {
				Oops ("cpy_u2k: error %d\n", -result);
         }
			else
         {
				Oops ("cpy_u2k: failed to copy %d bytes\n", result);
         }
		}
	}
  
	return (result);
#else
	memcpy_fromfs(dst, src, size);
  
	return 0;
#endif
}

int cpy_k2u(void *dst, const void* src, size_t size)
{
  
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)
	int		result = 0;

	if (size > 0)
	{
		if ((result = srr_safe_copy (dst, src, size, 1,
									 srr_copy_to_user)) != 0)
		{
			if (result < 0)
         {
				Oops ("cpy_k2u: error %d\n", -result);
         }
			else
         {
				Oops ("cpy_k2u: failed to copy %d bytes\n", result);
         }
		}
	}
  
	return (result);
#else
	memcpy_tofs(dst, src, size);
	
  return 0;
#endif
}

/*
 * memory xfer: get variable sized data from user space into a buffer
 */

char* str_cpy_u2k(const char* src, size_t size)
{
    char*         str_buffer;
  	
	if(src == 0 || size == 0){
    
		return 0;
  }

    /* Aditi Hatwalne <aditi.hatwalne@gmail.com> The 2.6.X kernel
     * series with pre-emption caused panics while running with large
     * number of processes even under uni-processor conditions.
	 * (CB) due to a race condition on a static buffer.  Static buffer
	 * removed.
     */

	str_buffer = kmalloc(size + 1, GFP_ATOMIC);
    if (!str_buffer){
        
        return 0; 
    }
	cpy_u2k(str_buffer, src, size);
	/* null terminate - don't just copy and hope its terminated correctly! */
	str_buffer[size] = '\0';
  
	return str_buffer;
}

void str_cpy_release(void* buffer)
{
	kfree(buffer);
}

void hexdump(void *data, size_t length)
{
   char *location = data;
   char curr;
   int i, remaining = length;
   char *output = (char *)kmalloc(HEX_STR_LEN, GFP_ATOMIC);

   while(remaining > 0)
   {
      if(remaining < BYTES_PER_LINE)
      {
       sprintf(output, "%.4x / ", ((int)location - (int)data));
      
       for(i = 0; i < remaining; i++)
       {
         curr = *(location + i);
         sprintf(output, "%s%x ", output, curr);
         if(i == BYTES_PER_DIVISION - 1)
           sprintf(output, "%s  ", output);
       }
       for(i = 0; i < (BYTES_PER_LINE - remaining); i++)
       {
         sprintf(output, "%s   ", output);
       }
       if(BYTES_PER_LINE - remaining >= BYTES_PER_DIVISION)
         sprintf(output, "%s  ", output);

       sprintf(output,"%s/ ", output);
       for(i = 0; i < remaining; i++)
       {
         curr = *(location + i);
         sprintf(output, "%s%c", output, curr);
       }
      }
      
      else
      {
       sprintf(output, "%.4x / ", ((int)location - (int)data));
       for(i = 0; i < BYTES_PER_LINE; i++)
       {
         curr = *(location + i);
         sprintf(output, "%s%x ", output, curr);
         if(i == BYTES_PER_DIVISION - 1)
           sprintf(output, "%s  ", output);
       }
       sprintf(output,"%s/ ", output);
       for(i = 0; i < BYTES_PER_LINE; i++)
       {
         curr = *(location + i);
         if(curr > 31 && curr != 127)
          sprintf(output, "%s%c", output, curr);
         else
          sprintf(output, "%s.", output);
       }
      }

      printk(KERN_DEBUG "%s\n", output);
      memset(output, '\0', HEX_STR_LEN);
      remaining -= BYTES_PER_LINE;
      location += BYTES_PER_LINE;
   }
   kfree(output);
}
  


